# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import json
import re
import select
import sys
import time
from typing import List, Optional

from .command import Command, IncrementalStyle
from .start import Start


class Persistent(Command):
    NAME = "persistent"

    def __init__(self, arguments, configuration, analysis_directory) -> None:
        super(Persistent, self).__init__(arguments, configuration, analysis_directory)

    def _run(self) -> None:
        arguments = self._arguments
        # pyre-fixme[16]: `Namespace` has no attribute `terminal`.
        arguments.terminal = False
        # pyre-fixme[16]: `Namespace` has no attribute `store_type_check_resolution`.
        arguments.store_type_check_resolution = False
        # pyre-fixme[16]: `Namespace` has no attribute `transitive`.
        arguments.incremental_style = IncrementalStyle.SHALLOW
        Start(arguments, self._configuration, self._analysis_directory).run()

        self._call_client(command=self.NAME, capture_output=False).check()

    def _flags(self) -> List[str]:
        flags = [
            "-log-identifier",
            '"{}"'.format(self._analysis_directory.get_root()),
            "-expected-binary-version",
            self._configuration.version_hash,
        ]
        if self._configuration.autocomplete:
            flags.append("-autocomplete")
        return flags

    @classmethod
    def _initialize_response(cls, request_id: int) -> str:
        response = json.dumps(
            {"jsonrpc": "2.0", "id": request_id, "result": {"capabilities": {}}},
            sort_keys=True,
        )
        return "Content-Length: {}\r\n\r\n{}\r\n".format(len(response), response)

    @classmethod
    def run_null_server(cls, timeout: Optional[int] = None) -> None:
        to_read, _, _ = select.select([sys.stdin], [], [], 3.0)
        request_id = 0
        if to_read:
            standard_input = to_read[0]
            # Read content of the form Content-Length:n\r\n\r\n{jsonmessage}
            line = standard_input.readline()
            match = re.match(r"Content-Length: (?P<bytes>[0-9]+)", line)
            if match:
                length = int(match.group("bytes"))
                standard_input.readline()
                serialized_json = standard_input.read(length)
                try:
                    parsed = json.loads(serialized_json)
                    request_id = parsed["id"]
                # This is a catch-all to ensure that the null server always
                # gets spawned.
                except Exception:
                    pass

        sys.stdout.write(cls._initialize_response(request_id))
        sys.stdout.flush()
        start_time = int(time.time())
        while True:
            if timeout is not None and timeout <= (int(time.time()) - start_time):
                break
            time.sleep(10)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from .analyze import Analyze as Analyze  # noqa
from .check import Check as Check  # noqa
from .color import Color as Color  # noqa
from .deobfuscate import Deobfuscate as Deobfuscate  # noqa
from .incremental import Incremental as Incremental  # noqa
from .infer import Infer as Infer  # noqa
from .initialize import Initialize as Initialize  # noqa
from .kill import Kill as Kill  # noqa
from .persistent import Persistent as Persistent  # noqa
from .query import Query as Query  # noqa
from .rage import Rage as Rage  # noqa
from .reporting import Reporting as Reporting  # noqa
from .restart import Restart as Restart  # noqa
from .start import Start  # noqa
from .statistics import Statistics as Statistics  # noqa
from .stop import Stop as Stop  # noqa


from .command import (  # noqa; noqa; noqa
    ClientException as ClientException,
    Command as Command,
    ExitCode as ExitCode,
    typeshed_search_path as typeshed_search_path,
    IncrementalStyle as IncrementalStyle,
)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import logging
import os
from typing import Any, Dict, List

from .commands import stop
from .configuration import CONFIGURATION_FILE
from .filesystem import AnalysisDirectory, is_parent
from .watchman_subscriber import Subscription, WatchmanSubscriber


LOG = logging.getLogger(__name__)


class ConfigurationMonitor(WatchmanSubscriber):
    """
        The ConfigurationMonitor watches only for .pyre_configuration(.local)
        files, and will kill the corresponding server and ProjectFileMonitor
        when a configuration changes.
        Logs are found in .pyre/configuration_monitor/configuration_monitor.log
        To kill a  monitor, get pid from
        .pyre/configuration_monitor/configuration_monitor.pid ; kill <pid>.
    """

    def __init__(
        self, arguments, configuration, analysis_directory: AnalysisDirectory
    ) -> None:
        super(ConfigurationMonitor, self).__init__(analysis_directory)
        self.arguments = arguments
        self.configuration = configuration
        self.analysis_directory = analysis_directory

    @property
    def _name(self) -> str:
        return "configuration_monitor"

    @property
    def _subscriptions(self) -> List[Subscription]:
        roots = self._watchman_client.query("watch-list")["roots"]
        names = ["pyre_monitor_{}".format(os.path.basename(root)) for root in roots]
        subscription = {
            "expression": [
                "allof",
                ["type", "f"],
                ["not", "empty"],
                [
                    "anyof",
                    ["suffix", "pyre_configuration.local"],
                    ["suffix", "pyre_configuration"],
                ],
            ],
            "fields": ["name"],
        }
        return [
            Subscription(root, name, subscription) for (root, name) in zip(roots, names)
        ]

    def _handle_response(self, response: Dict[str, Any]) -> None:
        LOG.info(
            "Update to configuration at %s",
            os.path.join(response["root"], ",".join(response["files"])),
        )
        absolute_path = [
            os.path.join(response["root"], file) for file in response["files"]
        ]

        # Find the path to the project configuration file and compare it with the
        # list of changed configuration files.
        project_configuration = os.path.join(response["root"], CONFIGURATION_FILE)
        if any((project_configuration == file) for file in absolute_path):
            LOG.info("Pyre configuration changed. Stopping pyre server.")
            stop.Stop(self.arguments, self.configuration, self.analysis_directory).run()
            # TODO(T54088045): Find all local pyre servers running underneath
            # and stop them.
        else:
            LOG.info("None of the changed paths correspond to pyre configuration.")

        if self.arguments.local_configuration:
            if any(
                is_parent(self.arguments.local_configuration, file)
                for file in absolute_path
            ):
                LOG.info("Local configuration changed. Stopping pyre server.")
                stop.Stop(
                    self.arguments, self.configuration, self.analysis_directory
                ).run()
            else:
                LOG.info(
                    "None of the changed paths correspond to the current local"
                    "configuration."
                )



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import logging
import os
import re
from typing import Dict, List, Optional

from .. import log
from .command import Command


LOG = logging.getLogger(__name__)


class Query(Command):
    NAME = "query"

    def _rewrite_paths(self, query: str) -> str:
        paths = re.findall(r"'[a-zA-Z_\-\.\/0-9]+\.py'", query)
        symbolic_link_mapping: Optional[Dict[str, str]] = None
        for path in paths:
            # Lazily compute the symbolic link mapping, as it can add over a second of
            # latency for large projects.
            if symbolic_link_mapping is None:
                symbolic_link_mapping = (
                    self._analysis_directory.compute_symbolic_links()
                )
            path = path[1:-1]
            absolute_path = os.path.abspath(path)
            if absolute_path in symbolic_link_mapping:
                query = query.replace(path, symbolic_link_mapping[absolute_path])
        return query

    def __init__(self, arguments, configuration, analysis_directory) -> None:
        super(Query, self).__init__(arguments, configuration, analysis_directory)
        self.query = self._rewrite_paths(arguments.query)

    def _flags(self) -> List[str]:
        return [self.query]

    def _run(self) -> None:
        result = self._call_client(command=self.NAME)
        result.check()
        log.stdout.write(result.output)



# Copyright (c) 2019-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import functools
import logging
import os
import signal
import sys
from multiprocessing import Event
from typing import Any, Dict, List, NamedTuple

from .filesystem import AnalysisDirectory, acquire_lock, remove_if_exists


LOG = logging.getLogger(__name__)  # type: logging.Logger


Subscription = NamedTuple(
    "Subscription", [("root", str), ("name", str), ("subscription", Dict[str, Any])]
)


class WatchmanSubscriber(object):
    def __init__(self, analysis_directory: AnalysisDirectory) -> None:
        self._base_path = os.path.join(
            analysis_directory.get_root(), ".pyre", self._name
        )  # type: str
        self._alive = True  # type: bool
        self._ready = Event()  # type: multiprocessing.synchronize.Event

    @property
    def _name(self) -> str:
        """
            A name to identify the subscriber. Used as the directory and file names
            for the log, lock, and pid files.
        """
        raise NotImplementedError

    @property
    def _subscriptions(self) -> List[Subscription]:
        """
            List of subscriptions
        """
        raise NotImplementedError

    def _handle_response(self, response: Dict[str, Any]) -> None:
        """
            Callback invoked when a message is received from watchman
        """
        raise NotImplementedError

    @property
    @functools.lru_cache(1)
    def _watchman_client(self) -> "pywatchman.client":  # noqa
        try:
            import pywatchman  # noqa

            # The client will block indefinitely when timeout is None.
            return pywatchman.client(timeout=None)
        except ImportError as exception:
            LOG.info("Not starting %s due to %s", self._name, str(exception))
            sys.exit(1)

    def _subscribe_to_watchman(self, subscription: Subscription) -> None:
        self._watchman_client.query(
            "subscribe", subscription.root, subscription.name, subscription.subscription
        )

    def _run(self) -> None:
        try:
            os.makedirs(self._base_path)
        except OSError:
            pass
        lock_path = os.path.join(self._base_path, "{}.lock".format(self._name))
        pid_path = os.path.join(self._base_path, "{}.pid".format(self._name))

        def cleanup() -> None:
            LOG.info("Cleaning up lock and pid files before exiting.")
            remove_if_exists(pid_path)
            remove_if_exists(lock_path)

        def interrupt_handler(_signal_number=None, _frame=None) -> None:
            LOG.info("Interrupt signal received.")
            cleanup()
            sys.exit(0)

        signal.signal(signal.SIGINT, interrupt_handler)

        # Die silently if unable to acquire the lock.
        with acquire_lock(lock_path, blocking=False):
            file_handler = logging.FileHandler(
                os.path.join(self._base_path, "%s.log" % self._name), mode="w"
            )
            file_handler.setFormatter(
                logging.Formatter("%(asctime)s %(levelname)s %(message)s")
            )
            LOG.addHandler(file_handler)

            with open(pid_path, "w+") as pid_file:
                pid_file.write(str(os.getpid()))

            for subscription in self._subscriptions:
                self._subscribe_to_watchman(subscription)

            connection = self._watchman_client.recvConn
            if not connection:
                LOG.error("Connection to Watchman for %s not found", self._name)
                sys.exit(1)

            while self._alive:
                # This call is blocking, which prevents this loop from burning CPU.
                response = connection.receive()
                try:
                    if response["is_fresh_instance"]:
                        LOG.info(
                            "Ignoring initial watchman message for %s", response["root"]
                        )
                    else:
                        self._handle_response(response)
                    self._ready.set()  # At least one message has been received.
                except KeyError:
                    pass

            cleanup()

    def daemonize(self) -> None:
        """We double-fork here to detach the daemon process from the parent.
           If we were to just fork the child as a daemon, we'd have to worry about the
           parent process exiting zombifying the daemon."""
        if os.fork() == 0:
            pid = os.fork()
            if pid == 0:
                try:
                    LOG.propagate = False
                    # Closing the sys.stdout and stderr file descriptors here causes
                    # the program to crash when attempting to log.
                    os.close(sys.stdout.fileno())
                    os.close(sys.stderr.fileno())
                    self._run()
                    sys.exit(0)
                except Exception as exception:
                    LOG.info("Not running %s due to %s", self._name, str(exception))
                    sys.exit(1)
            else:
                sys.exit(0)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import json
import logging
import multiprocessing
import os
import platform
import subprocess
import sys
import time
import traceback
from argparse import Namespace
from typing import Any, Dict, Optional

from . import buck
from .exceptions import EnvironmentException
from .filesystem import (  # noqa
    AnalysisDirectory,
    SharedAnalysisDirectory,
    find_root,
    translate_path,
    translate_paths,
)


LOG = logging.getLogger(__name__)


CONFIGURATION_FILE = ".pyre_configuration"
BINARY_NAME = "pyre.bin"
CLIENT_NAME = "pyre-client"


def assert_readable_directory(directory: str) -> None:
    if not os.path.isdir(directory):
        raise EnvironmentException("{} is not a valid directory.".format(directory))
    if not os.access(directory, os.R_OK):
        raise EnvironmentException("{} is not a readable directory.".format(directory))


def assert_writable_directory(directory: str) -> None:
    if not os.path.isdir(directory):
        raise EnvironmentException("{} is not a valid directory.".format(directory))
    if not os.access(directory, os.W_OK):
        raise EnvironmentException("{} is not a writable directory.".format(directory))


def readable_directory(directory: str) -> str:
    assert_readable_directory(directory)
    return directory


def is_capable_terminal(file=sys.stderr) -> bool:
    """
    Determine whether we are connected to a capable terminal.
    """
    if not os.isatty(file.fileno()):
        return False
    terminal = os.getenv("TERM", "dumb")
    # Hardcoded list of non-capable terminals.
    return terminal not in ["dumb", "emacs"]


def get_binary_version(configuration) -> str:
    override = os.getenv("PYRE_BINARY")
    if override:
        return "override: {}".format(override)

    configured = configuration.version_hash
    if configured:
        return configured

    return "No version set"


def get_binary_version_from_file(local_path: Optional[str]) -> str:
    override = os.getenv("PYRE_BINARY")
    if override:
        return "override: {}".format(override)

    def read_version(configuration_path: str) -> Optional[str]:
        with open(configuration_path) as file:
            configuration_contents = file.read()
            return json.loads(configuration_contents).pop("version", None)

    version = None  # type: Optional[str]
    try:
        # Get local configuration version
        if local_path:
            local_configuration = os.path.join(
                local_path, CONFIGURATION_FILE + ".local"
            )
            version = read_version(local_configuration)

        # Get configuration version
        if not version:
            version = read_version(CONFIGURATION_FILE)
    except Exception:
        pass
    # pyre-fixme[7]: Expected `str` but got `Optional[str]`.
    return "No version set" if not version else version


def switch_root(arguments) -> None:
    arguments.original_directory = os.getcwd()
    local_root = find_root(arguments.original_directory, CONFIGURATION_FILE + ".local")
    global_root = find_root(arguments.original_directory, CONFIGURATION_FILE)

    # Check for illegal nested local configuration.
    if local_root:
        parent_local_root = find_root(
            os.path.dirname(local_root), CONFIGURATION_FILE + ".local"
        )
        if parent_local_root:
            raise EnvironmentException(
                "Local configuration is nested under another local configuration at "
                "`{}`. Please combine the sources into a single configuration or split "
                "the parent configuration to avoid inconsistent errors.".format(
                    parent_local_root
                )
            )

    # If the global configuration root is deeper than local configuration, ignore local.
    if global_root and local_root and global_root.startswith(local_root):
        local_root = None
    if local_root and arguments.local_configuration is None:
        arguments.local_configuration = local_root

    root = global_root or arguments.original_directory
    os.chdir(root)
    arguments.current_directory = root


def translate_arguments(commands, arguments):
    root = arguments.original_directory

    if arguments.command in [commands.Analyze]:
        if arguments.taint_models_path:
            arguments.taint_models_path = [
                translate_path(root, path) for path in arguments.taint_models_path
            ]

    if arguments.logger:
        arguments.logger = translate_path(root, arguments.logger)


def _buck_target_count(arguments, configuration):
    if arguments.source_directories or arguments.targets:
        return len(set(arguments.targets or []))
    else:
        return len(set(configuration.targets or []))


def _resolve_filter_paths(arguments, configuration):
    filter_paths = []
    if arguments.source_directories or arguments.targets:
        if arguments.source_directories:
            filter_paths += arguments.source_directories
        if arguments.targets:
            filter_paths += [
                buck.presumed_target_root(target) for target in arguments.targets
            ]
    else:
        local_configuration_root = configuration.local_configuration_root
        if local_configuration_root:
            filter_paths = [local_configuration_root]
    return translate_paths(filter_paths, arguments.original_directory)


def resolve_analysis_directory(
    arguments, commands, configuration, isolate: bool = False
) -> AnalysisDirectory:
    # Only read from the configuration if no explicit targets are passed in.
    if not arguments.source_directories and not arguments.targets:
        source_directories = configuration.source_directories
        targets = configuration.targets
    else:
        source_directories = arguments.source_directories or []
        targets = arguments.targets or []
        if targets:
            configuration_name = ".pyre_configuration.local"
            command = "pyre init --local"
        else:
            configuration_name = ".pyre_configuration"
            command = "pyre init"
        LOG.warning(
            "Setting up a `%s` with `%s` may reduce overhead.",
            configuration_name,
            command,
        )

    if arguments.filter_directory:
        filter_paths = [arguments.filter_directory]
    else:
        filter_paths = _resolve_filter_paths(arguments, configuration)

    local_configuration_root = configuration.local_configuration_root
    if local_configuration_root:
        local_configuration_root = os.path.relpath(
            local_configuration_root, arguments.current_directory
        )

    use_buck_builder = (
        arguments.use_buck_builder or configuration.use_buck_builder
    ) and not arguments.use_legacy_builder

    if len(source_directories) == 1 and len(targets) == 0:
        analysis_directory = AnalysisDirectory(
            source_directories.pop(),
            filter_paths=filter_paths,
            search_path=configuration.search_path,
        )
    else:
        build = arguments.build or arguments.command in [
            commands.Analyze,
            commands.Check,
            commands.Restart,
        ]
        buck_builder = buck.SimpleBuckBuilder(build=build)
        if use_buck_builder:
            buck_root = buck.find_buck_root(os.getcwd())
            if not buck_root:
                raise EnvironmentException(
                    "No Buck configuration at `{}` or any of its ancestors.".format(
                        os.getcwd()
                    )
                )
            buck_builder = buck.FastBuckBuilder(
                buck_root=buck_root,
                buck_builder_binary=arguments.buck_builder_binary,
                buck_builder_target=arguments.buck_builder_target,
                debug_mode=arguments.buck_builder_debug,
            )
        else:
            buck_builder = buck.SimpleBuckBuilder(build=build)

        analysis_directory = SharedAnalysisDirectory(
            source_directories=source_directories,
            targets=targets,
            buck_builder=buck_builder,
            original_directory=arguments.original_directory,
            filter_paths=filter_paths,
            local_configuration_root=local_configuration_root,
            extensions=configuration.extensions,
            search_path=configuration.search_path,
            isolate=isolate,
        )
    return analysis_directory


def number_of_workers() -> int:
    try:
        return max(multiprocessing.cpu_count() - 4, 1)
    except NotImplementedError:
        return 4


def log_statistics(
    category: str,
    arguments: Optional[Namespace] = None,
    # this is typed as a Any because configuration imports __init__
    configuration: Optional[Any] = None,
    integers: Optional[Dict[str, int]] = None,
    normals: Optional[Dict[str, str]] = None,
    logger: Optional[str] = None,
) -> None:
    integers = integers or {}
    if "time" not in integers:
        integers["time"] = int(time.time())
    normals = normals or {}
    if configuration:
        normals = {
            **normals,
            "version": configuration.version_hash,
        }  # type: Dict[str, str]
        if not logger:
            logger = configuration.logger
    if not logger:
        raise ValueError("Logger must either be given or in configuration")
    if arguments:
        normals = {**normals, "arguments": str(arguments)}
    try:
        statistics = {
            "int": integers,
            "normal": {
                **normals,
                "command_line": " ".join(sys.argv),
                "host": platform.node() or "",
                "platform": platform.system() or "",
                "user": os.getenv("USER", ""),
            },
        }
        statistics = json.dumps(statistics).encode("ascii", "strict")
        subprocess.run([logger, category], input=statistics)
    except Exception:
        LOG.warning("Unable to log using `%s`", logger)
        LOG.info(traceback.format_exc())


def _find_directory_upwards(base: str, target: str) -> Optional[str]:
    """
    Walk directories upwards from base, until the root directory is
    reached. At each step, check if the target directory exist, and return
    it if found. Return None if the search is unsuccessful.
    """
    while True:
        step = os.path.join(base, target)
        LOG.debug("Trying with: `%s`", step)
        if os.path.isdir(step):
            return step
        parent_directory = os.path.dirname(base)
        if parent_directory == base:
            # We have reached the root.
            break
        base = parent_directory
    return None


def find_typeshed() -> Optional[str]:
    override = os.getenv("PYRE_TYPESHED")
    if override:
        return override

    current_directory = os.path.dirname(os.path.realpath(__file__))

    # Prefer the typeshed we bundled ourselves (if any) to the one
    # from the environment.
    bundled_typeshed = _find_directory_upwards(
        current_directory, "pyre_check/typeshed/"
    )
    if bundled_typeshed:
        return bundled_typeshed

    try:
        import typeshed  # pyre-fixme: Can't find module import typeshed

        return typeshed.typeshed
    except ImportError:
        LOG.debug("`import typeshed` failed, attempting a manual lookup")

    # This is a terrible, terrible hack.
    return _find_directory_upwards(current_directory, "typeshed/")



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import atexit
import logging
import os
import subprocess
import sys
from typing import IO, List, cast

from ..project_files_monitor import MonitorException, ProjectFilesMonitor
from .command import (
    ClientException,
    ExitCode,
    IncrementalStyle,
    State,
    typeshed_search_path,
)
from .reporting import Reporting
from .start import Start


LOG = logging.getLogger(__name__)


class Incremental(Reporting):
    NAME = "incremental"

    def __init__(self, arguments, configuration, analysis_directory) -> None:
        super(Incremental, self).__init__(arguments, configuration, analysis_directory)
        self._nonblocking = arguments.nonblocking  # type: bool
        self._incremental_style = arguments.incremental_style  # type: bool

    def _run(self) -> None:
        if self._state() == State.DEAD:
            LOG.warning("Starting server at `%s`.", self._analysis_directory.get_root())
            arguments = self._arguments
            # pyre-fixme[16]: `Namespace` has no attribute `terminal`.
            arguments.terminal = False
            # pyre-fixme[16]: `Namespace` has no attribute
            #  `store_type_check_resolution`.
            arguments.store_type_check_resolution = False
            # pyre-fixme[16]: `Namespace` has no attribute `no_watchman`.
            arguments.no_watchman = False
            exit_code = (
                Start(arguments, self._configuration, self._analysis_directory)
                .run()
                .exit_code()
            )
            if exit_code != ExitCode.SUCCESS:
                self._exit_code = ExitCode.FAILURE
                return
        else:
            self._refresh_file_monitor()

        if self._state() != State.DEAD:
            LOG.info("Waiting for server...")

        result = self._call_client(command=self.NAME)

        try:
            result.check()
            errors = self._get_errors(result)
            self._print(errors)

            if errors:
                self._exit_code = ExitCode.FOUND_ERRORS
        except ClientException as exception:
            LOG.error("Error while waiting for server: %s", str(exception))
            LOG.error("Run `pyre restart` in order to restart the server.")
            self._exit_code = ExitCode.FAILURE

    def _flags(self) -> List[str]:
        flags = super()._flags()
        flags.extend(["-expected-binary-version", self._configuration.version_hash])

        search_path = self._configuration.search_path + typeshed_search_path(
            self._configuration.typeshed
        )
        if search_path:
            flags.extend(["-search-path", ",".join(search_path)])

        excludes = self._configuration.excludes
        for exclude in excludes:
            flags.extend(["-exclude", exclude])

        extensions = self._configuration.extensions
        for extension in extensions:
            flags.extend(["-extension", extension])

        if self._nonblocking:
            flags.append("-nonblocking")

        if self._incremental_style == IncrementalStyle.TRANSITIVE:
            flags.append("-transitive")

        return flags

    def _read_stderr(self, _stream) -> None:
        stderr_file = os.path.join(
            self._analysis_directory.get_root(), ".pyre/server/server.stdout"
        )
        with subprocess.Popen(
            ["tail", "--follow", "--lines=0", stderr_file],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        ) as stderr_tail:
            atexit.register(stderr_tail.terminate)
            super(Incremental, self)._read_stderr(cast(IO[bytes], stderr_tail.stdout))

    def _refresh_file_monitor(self) -> None:
        if not ProjectFilesMonitor.is_alive(self._analysis_directory.get_root()):
            LOG.info("File monitor is not running.")
            try:
                ProjectFilesMonitor(
                    self._arguments, self._configuration, self._analysis_directory
                ).daemonize()
                LOG.info("Restarted file monitor.")
            except MonitorException as exception:
                LOG.warning("Failed to restart file monitor: %s", exception)



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import argparse
import copy
import io
import logging
import os
import re
import sys
import threading
import time
from typing import List, Optional, Sequence  # noqa

from . import filesystem


LOG = logging.getLogger(__name__)  # type: logging.Logger
PERFORMANCE = 15  # type: int
PROMPT = 50  # type: int
SUCCESS = 60  # type: int


stdout = io.StringIO(newline="")  # type: io.StringIO


class Color:
    YELLOW = "\033[33m"  # type: str
    RED = "\033[31m"  # type: str
    GREEN = "\033[32m"  # type: str


class Format:
    BOLD = "\033[1m"  # type: str

    CLEAR_LINE = "\x1b[0G\x1b[K"  # type: str
    CLEAR = "\033[0m"  # type: str
    TRUNCATE_OVERFLOW = "\033[?7l"  # type: str
    WRAP_OVERFLOW = "\033[?7h"  # type: str
    NEWLINE = "\n"  # type: str

    CURSOR_UP_LINE = "\x1b[1A"  # type: str
    HIDE_CURSOR = "\x1b[?25l"  # type: str
    SHOW_CURSOR = "\x1b[?25h"  # type: str


class Character:
    LAMBDA = "ƛ"  # type: str


class SectionFormatter(logging.Formatter):
    def __init__(self) -> None:
        super(SectionFormatter, self).__init__("%(asctime)s %(levelname)s %(message)s")

    def format(self, record: logging.LogRecord) -> str:
        formatted = super(SectionFormatter, self).format(record)
        return re.sub(r"DEBUG \[(.*)\]", r"\1", formatted)


class TimedStreamHandler(logging.StreamHandler):
    THRESHOLD = 0.5  # type: float
    LINE_BREAKING_LEVELS = ["ERROR", "WARNING", "SUCCESS"]  # type: Sequence[str]

    _terminate = False  # type: bool
    _last_update = 0.0  # type: float

    def __init__(self) -> None:
        super(TimedStreamHandler, self).__init__()
        self.setFormatter(logging.Formatter("%(message)s"))
        self.terminator = ""  # type: str
        self.setLevel(logging.INFO)

        self._record = None  # type: Optional[logging.LogRecord]
        self._last_record = None  # type: Optional[logging.LogRecord]
        self._active_lines = 0  # type: int

        # Preamble preparing terminal.
        sys.stderr.write(
            Format.NEWLINE
            + Format.CLEAR_LINE
            + Format.CURSOR_UP_LINE
            + Format.HIDE_CURSOR
        )

        thread = threading.Thread(target=self._thread)
        thread.daemon = True
        thread.start()

    def clear_lines(self) -> str:
        if self._active_lines == 0:
            return ""
        return Format.CLEAR_LINE + "".join(
            [
                Format.CURSOR_UP_LINE + Format.CLEAR_LINE
                for n in range(self._active_lines - 1)
            ]
        )

    def emit(self, record: logging.LogRecord, age: Optional[float] = None) -> None:
        self._last_record = record
        suffix = ""
        color = ""
        active_lines = record.msg.count("\n") + 1
        if record.levelname in self.LINE_BREAKING_LEVELS:
            record.msg += "\n"

        if record.levelname == "ERROR":
            color = Color.RED
            self._record = None
            active_lines = 0
        elif record.levelname == "WARNING":
            color = Color.YELLOW
            self._record = None
            active_lines = 0
        elif record.levelname == "PROMPT":
            color = Color.YELLOW
            self._record = None
            active_lines = 0
        elif record.levelname == "SUCCESS":
            self._record = None
            active_lines = 0
        elif age:
            if age > 10:
                color = Color.YELLOW
            if age > 30:
                color = Color.RED
            suffix = " {}[{:.1f}s]{}".format(
                color if color else "", age, Format.CLEAR if color else ""
            )
        else:
            self._record = record
            self._last_update = time.time()

        timed_record = copy.copy(record)
        timed_record.msg = (
            "{clear_line}{color} {cursor}{clear} " "{truncate}{message}{suffix}"
        ).format(
            clear_line=self.clear_lines(),
            color=color,
            cursor=Character.LAMBDA,
            clear=Format.CLEAR,
            truncate=Format.TRUNCATE_OVERFLOW,
            message=record.msg,
            suffix=suffix,
        )
        self._active_lines = active_lines
        super(TimedStreamHandler, self).emit(timed_record)

    def _thread(self) -> None:
        while not self._terminate:
            record = self._record
            if record:
                age = time.time() - self._last_update
                if age > self.THRESHOLD:
                    self.emit(record, age)
            time.sleep(0.1)

    def terminate(self) -> None:
        last_record = self._last_record
        if last_record and last_record.levelname not in self.LINE_BREAKING_LEVELS:
            sys.stderr.write("\n")

        # Reset terminal.
        sys.stderr.write(Format.WRAP_OVERFLOW + Format.SHOW_CURSOR)
        sys.stderr.flush()
        self._terminate = True


def initialize(arguments: argparse.Namespace) -> None:
    if arguments.noninteractive:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(SectionFormatter())
        stream_handler.setLevel(logging.DEBUG)
        # pyre-fixme[16]: `Namespace` has no attribute `timed_stream_handler`.
        arguments.timed_stream_handler = None
    else:
        stream_handler = TimedStreamHandler()
        arguments.timed_stream_handler = stream_handler

    handlers = [stream_handler]  # type: List[logging.Handler]

    if not arguments.noninteractive:
        pyre_directory = filesystem.make_pyre_directory()
        file_handler = logging.FileHandler(os.path.join(pyre_directory, "pyre.stderr"))
        file_handler.setFormatter(SectionFormatter())
        file_handler.setLevel(logging.DEBUG)
        handlers.append(file_handler)
    logging.addLevelName(PERFORMANCE, "PERFORMANCE")
    logging.addLevelName(PROMPT, "PROMPT")
    logging.addLevelName(SUCCESS, "SUCCESS")
    logging.basicConfig(level=logging.DEBUG, handlers=handlers)


def cleanup(arguments: argparse.Namespace) -> None:
    if arguments.timed_stream_handler:
        arguments.timed_stream_handler.terminate()

    output = stdout.getvalue()
    if output:
        sys.stdout.write(output + "\n")


class Buffer:
    THRESHOLD = 0.1  # type: float

    _flushed = False  # type: bool

    def __init__(self, section: str, data: List[str]) -> None:
        self._section = section  # type: str
        self._data = data  # type: List[str]
        self._lock = threading.RLock()  # type: threading.RLock
        thread = threading.Thread(target=self._thread)
        thread.daemon = True
        thread.start()

    def append(self, line: str) -> None:
        self._data.append(line)

    def flush(self) -> None:
        with self._lock:
            if self._flushed is True:
                return
            self._flushed = True
        message = "\n".join(self._data)
        if self._section == "ERROR":
            LOG.error(message)
        elif self._section == "INFO":
            LOG.info(message)
        elif self._section == "DUMP":
            LOG.warning(message)
        elif self._section == "WARNING":
            LOG.warning(message)
        elif self._section == "PROGRESS":
            LOG.info(message)
        elif self._section == "PARSER":
            LOG.error(message)
        else:
            LOG.debug("[%s] %s", self._section, message)

    def _thread(self) -> None:
        time.sleep(self.THRESHOLD)
        with self._lock:
            if not self._flushed:
                self.flush()


def get_yes_no_input(prompt: str) -> bool:
    choice = get_input(prompt, suffix=" [Y/n] ")
    return choice.lower() in ["", "y", "ye", "yes"]


def get_optional_input(prompt: str, default: str) -> str:
    result = get_input(prompt, suffix=" (Default: `{}`): ".format(default))
    if result == "":
        return default
    return result


def get_input(prompt: str, suffix: str = "") -> str:
    LOG.log(PROMPT, prompt + suffix)
    return input().strip()



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import argparse
import logging
import os
import shutil
import sys
import time
import traceback
from typing import Type  # noqa

from . import (
    assert_writable_directory,
    buck,
    commands,
    get_binary_version_from_file,
    is_capable_terminal,
    log,
    log_statistics,
    readable_directory,
    resolve_analysis_directory,
    switch_root,
    translate_arguments,
)
from .commands import Command, ExitCode, IncrementalStyle, reporting  # noqa
from .configuration import Configuration
from .exceptions import EnvironmentException
from .filesystem import AnalysisDirectory
from .version import __version__


LOG = logging.getLogger(__name__)  # type: logging.Logger


def main() -> int:
    def executable_file(file_path: str) -> str:
        if not os.path.isfile(file_path):
            raise EnvironmentException("%s is not a valid file" % file_path)
        if not os.access(file_path, os.X_OK):
            raise EnvironmentException("%s is not an executable file" % file_path)
        return file_path

    def writable_directory(path: str) -> str:
        # Create the directory if it does not exist.
        try:
            os.makedirs(path)
        except FileExistsError:
            pass
        assert_writable_directory(path)
        return path

    def file_exists(path: str) -> str:
        if not os.path.exists(path):
            raise argparse.ArgumentTypeError("ERROR: " + str(path) + " does not exist")
        return path

    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        formatter_class=argparse.RawTextHelpFormatter,
        epilog="environment variables:"
        "\n   `PYRE_BINARY` overrides the pyre binary used."
        "\n   `PYRE_VERSION_HASH` overrides the pyre version set in the "
        "configuration files.",
    )

    parser.add_argument(
        "-l", "--local-configuration", type=str, help="Use a local configuration"
    )

    parser.add_argument(
        "--version",
        action="store_true",
        help="Print the client and binary versions of Pyre.",
    )

    parser.add_argument("--debug", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--sequential", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--strict", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--additional-check", action="append", help=argparse.SUPPRESS)

    parser.add_argument(
        "--show-error-traces",
        action="store_true",
        help="Display errors trace information",
    )

    # Logging.
    parser.add_argument(
        "--output",
        choices=[commands.reporting.TEXT, commands.reporting.JSON],
        default=commands.reporting.TEXT,
        help="How to format output",
    )
    parser.add_argument("--verbose", action="store_true", help="Enable verbose logging")
    parser.add_argument(
        "--enable-profiling", action="store_true", help=argparse.SUPPRESS
    )
    parser.add_argument(
        "-n",
        "--noninteractive",
        action="store_true",
        help="Disable interactive logging",
    )
    parser.add_argument(
        "--hide-parse-errors",
        action="store_true",
        help="Hide detailed information about parse errors",
    )
    parser.add_argument(
        "--show-parse-errors",
        action="store_true",
        help="[DEPRECATED] Show detailed information about parse errors",
    )
    parser.add_argument(
        "--logging-sections", help=argparse.SUPPRESS  # Enable sectional logging.
    )
    parser.add_argument(
        "--log-identifier",
        default="",
        help=argparse.SUPPRESS,  # Add given identifier to logged samples.
    )
    parser.add_argument(
        "--logger", help=argparse.SUPPRESS  # Specify custom logging binary.
    )

    # Link tree determination.
    buck_arguments = parser.add_argument_group("buck")
    buck_arguments.add_argument(
        "--target", action="append", dest="targets", help="The buck target to check"
    )
    buck_arguments.add_argument(
        "--build",
        action="store_true",
        help="Freshly build all the necessary artifacts.",
    )
    buck_arguments.add_argument(
        "--use-buck-builder",
        action="store_true",
        help="Use Pyre's experimental builder for Buck projects.",
    )
    buck_arguments.add_argument(
        "--use-legacy-builder",
        action="store_true",
        help="Use Pyre's legacy builder for Buck projects.",
    )
    buck_arguments.add_argument(
        "--buck-builder-debug", action="store_true", help=argparse.SUPPRESS
    )

    source_directories = parser.add_argument_group("source-directories")
    source_directories.add_argument(
        "--source-directory",
        action="append",
        dest="source_directories",
        help="The source directory to check",
        type=os.path.abspath,
    )
    source_directories.add_argument(
        "--filter-directory", help=argparse.SUPPRESS  # override filter directory
    )

    parser.add_argument(
        "--use-global-shared-analysis-directory",
        action="store_true",
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        "--no-saved-state",
        action="store_true",
        help="Don't attempt to load Pyre from a saved state.",
    )

    # Handling of search path
    parser.add_argument(
        "--search-path",
        action="append",
        default=[],
        type=readable_directory,
        help="Add an additional directory of modules and stubs to include"
        " in the type environment",
    )
    parser.add_argument(
        "--preserve-pythonpath",
        action="store_true",
        default=False,
        help="Preserve the value of the PYTHONPATH environment variable and "
        "inherit the current python environment's search path",
    )

    parser.add_argument(
        "--binary",
        default=None,
        type=executable_file,
        help="Location of the pyre binary",
    )

    parser.add_argument(
        "--buck-builder-binary",
        default=None,
        help="Location of the buck builder binary",
    )
    parser.add_argument("--buck-builder-target", default=None, help=argparse.SUPPRESS)

    parser.add_argument(
        "--exclude",
        action="append",
        default=[],
        help="Exclude files and directories matching this regexp from parsing",
    )

    # Typeshed stubs location
    parser.add_argument(
        "--typeshed",
        default=None,
        type=readable_directory,
        help="Location of the typeshed stubs",
    )
    parser.add_argument(
        "--save-initial-state-to",
        default=None,
        help="Path to serialize pyre's initial state to.",
    )
    parser.add_argument(
        "--load-initial-state-from", default=None, type=str, help=argparse.SUPPRESS
    )
    parser.add_argument(
        "--changed-files-path", default=None, type=str, help=argparse.SUPPRESS
    )
    parser.add_argument(
        "--saved-state-project", default=None, type=str, help=argparse.SUPPRESS
    )

    # Subcommands.
    parsed_commands = parser.add_subparsers(
        metavar="{analyze, check, color, kill, incremental, initialize (init), "
        "query, rage, restart, start, stop}",
        help="""
        The pyre command to run; defaults to `incremental`.
        Run `pyre command --help` for documentation on a specific command.
        """,
    )

    incremental_help = """
    Connects to a running Pyre server and returns the current type errors for your
    project. If no server exists for your projects, starts a new one. Running `pyre`
    implicitly runs `pyre incremental`.

    By default, incremental checks ensure that all dependencies of changed files are
    analyzed before returning results. If you'd like to get partial type checking
    results eagerly, you can run `pyre incremental --nonblocking`.
    """
    incremental = parsed_commands.add_parser(
        commands.Incremental.NAME, epilog=incremental_help
    )
    incremental.set_defaults(command=commands.Incremental)
    incremental.add_argument(
        "--nonblocking",
        action="store_true",
        help=(
            "Ask the server to return partial results immediately, "
            "even if analysis is still in progress."
        ),
    )
    incremental.add_argument(
        "--incremental-style",
        type=IncrementalStyle,
        choices=list(IncrementalStyle),
        default=IncrementalStyle.SHALLOW,
        help="How to approach doing incremental checks.",
    )
    rage = parsed_commands.add_parser(
        commands.Rage.NAME,
        epilog="""
        Collects troubleshooting diagnostics for Pyre, and writes this information to
        the terminal.
        """,
    )
    rage.set_defaults(command=commands.Rage)

    check = parsed_commands.add_parser(
        commands.Check.NAME,
        epilog="""
      Runs a one-time check of a project without initializing a type check server.
    """,
    )
    check.set_defaults(command=commands.Check)

    color = parsed_commands.add_parser(commands.Color.NAME)
    color.add_argument("path")
    color.set_defaults(command=commands.Color)

    deobfuscate = parsed_commands.add_parser(commands.Deobfuscate.NAME)

    deobfuscate.set_defaults(command=commands.Deobfuscate)

    analyze = parsed_commands.add_parser(commands.Analyze.NAME)
    analyze.set_defaults(command=commands.Analyze)
    analyze.add_argument(
        "analysis", nargs="?", default="taint", help="Type of analysis to run: {taint}"
    )
    analyze.add_argument(
        "--taint-models-path",
        action="append",
        default=[],
        type=readable_directory,
        help="Location of taint models",
    )
    analyze.add_argument(
        "--no-verify",
        action="store_true",
        help="Do not verify models for the taint analysis.",
    )
    analyze.add_argument(
        "--save-results-to",
        default=None,
        type=writable_directory,
        help="Directory to write analysis results to.",
    )
    analyze.add_argument("--dump-call-graph", action="store_true")

    persistent = parsed_commands.add_parser(
        commands.Persistent.NAME,
        epilog="""
        Entry point for IDE integration to Pyre. Communicates with a
        Pyre server using the Language Server Protocol, accepts input from stdin and
        writing diagnostics and responses from the Pyre server to stdout.
        """,
    )
    persistent.add_argument(
        "--no-watchman",
        action="store_true",
        help="Do not spawn a watchman client in the background.",
    )
    persistent.set_defaults(command=commands.Persistent, noninteractive=True)

    start = parsed_commands.add_parser(
        commands.Start.NAME, epilog="Starts a pyre server as a daemon."
    )
    start.add_argument(
        "--terminal", action="store_true", help="Run the server in the terminal."
    )
    start.add_argument(
        "--store-type-check-resolution",
        action="store_true",
        help="Store extra information for `types` queries.",
    )
    start.add_argument(
        "--no-watchman",
        action="store_true",
        help="Do not spawn a watchman client in the background.",
    )
    start.add_argument(
        "--incremental-style",
        type=IncrementalStyle,
        choices=list(IncrementalStyle),
        default=IncrementalStyle.SHALLOW,
        help="How to approach doing incremental checks.",
    )
    start.set_defaults(command=commands.Start)

    stop = parsed_commands.add_parser(
        commands.Stop.NAME, epilog="Signals the Pyre server to stop."
    )
    stop.set_defaults(command=commands.Stop)

    restart = parsed_commands.add_parser(
        commands.Restart.NAME,
        epilog="Restarts a server. Equivalent to `pyre stop && pyre start`.",
    )
    restart.add_argument(
        "--terminal", action="store_true", help="Run the server in the terminal."
    )
    restart.add_argument(
        "--store-type-check-resolution",
        action="store_true",
        help="Store extra information for `types` queries.",
    )
    restart.add_argument(
        "--no-watchman",
        action="store_true",
        help="Do not spawn a watchman client in the background.",
    )
    restart.add_argument(
        "--incremental-style",
        type=IncrementalStyle,
        choices=list(IncrementalStyle),
        default=IncrementalStyle.SHALLOW,
        help="How to approach doing incremental checks.",
    )
    restart.set_defaults(command=commands.Restart)

    kill = parsed_commands.add_parser(commands.Kill.NAME)
    kill.add_argument(
        "--with-fire", action="store_true", help="Adds emphasis to the command."
    )
    kill.set_defaults(command=commands.Kill)

    initialize = parsed_commands.add_parser(commands.Initialize.NAME, aliases=["init"])
    initialize.add_argument(
        "--local",
        action="store_true",
        help="Initializes a local configuration in a project subdirectory.",
    )
    initialize.set_defaults(command=commands.Initialize)

    query_message = """
    `https://pyre-check.org/docs/querying-pyre.html` contains examples and documentation
    for this command, which queries a running pyre server for type, function and
    attribute information.

    To get a full list of queries, you can run `pyre query help`.
    """
    query = parsed_commands.add_parser(commands.Query.NAME, epilog=query_message)
    query_argument_message = """
    `pyre query help` will give a full list of available queries for the running Pyre.
     Example: `pyre query "superclasses(int)"`.
    """
    query.add_argument("query", help=query_argument_message)
    query.set_defaults(command=commands.Query)

    infer = parsed_commands.add_parser(commands.Infer.NAME)
    infer.add_argument(
        "-p",
        "--print-only",
        action="store_true",
        help="Print raw JSON errors to standard output, "
        + "without converting to stubs or annnotating.",
    )
    infer.add_argument(
        "-f",
        "--full-only",
        action="store_true",
        help="Only output fully annotated functions. Requires infer flag.",
    )
    infer.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="Recursively run infer until no new annotations are generated."
        + " Requires infer flag.",
    )
    infer.add_argument(
        "-i",
        "--in-place",
        nargs="*",
        metavar="path",
        type=file_exists,
        help="Add annotations to functions in selected paths."
        + " Takes a set of files and folders to add annotations to."
        + " If no paths are given, all functions are annotated."
        + " WARNING: Modifies original files and requires infer flag and retype",
    )
    infer.add_argument(
        "--json",
        action="store_true",
        help="Accept JSON input instead of running full check.",
    )
    infer.add_argument(
        "--annotate-from-existing-stubs",
        action="store_true",
        help="Add annotations from existing stubs.",
    )
    infer.add_argument(
        "--debug-infer",
        action="store_true",
        help="Print error message when file fails to annotate.",
    )
    infer.set_defaults(command=commands.Infer)

    statistics = parsed_commands.add_parser(commands.Statistics.NAME)
    statistics.add_argument(
        "filter_paths",
        nargs="*",
        type=file_exists,
        help="Source path(s) to gather metrics for.",
    )
    statistics.set_defaults(command=commands.Statistics)

    arguments = parser.parse_args()

    if not hasattr(arguments, "command"):
        if shutil.which("watchman"):
            # pyre-fixme[16]: `Namespace` has no attribute `command`.
            arguments.command = commands.Incremental
            # pyre-fixme[16]: `Namespace` has no attribute `nonblocking`.
            arguments.nonblocking = False
            # pyre-fixme[16]: `Namespace` has no attribute `transitive`.
            arguments.incremental_style = IncrementalStyle.SHALLOW
        else:
            watchman_link = "https://facebook.github.io/watchman/docs/install.html"
            LOG.warning(
                "No watchman binary found. \n"
                "To enable pyre incremental, "
                "you can install watchman: {}".format(watchman_link)
            )
            LOG.warning("Defaulting to non-incremental check.")
            arguments.command = commands.Check

    configuration = None
    analysis_directory = None
    # Having this as a fails-by-default helps flag unexpected exit
    # from exception flows.
    exit_code = ExitCode.FAILURE
    start = time.time()
    try:
        # pyre-fixme[16]: `Namespace` has no attribute `capable_terminal`.
        arguments.capable_terminal = is_capable_terminal()
        if arguments.debug or not arguments.capable_terminal:
            # pyre-fixme[16]: `Namespace` has no attribute `noninteractive`.
            arguments.noninteractive = True

        switch_root(arguments)
        translate_arguments(commands, arguments)
        log.initialize(arguments)

        if arguments.command in [commands.Initialize]:
            analysis_directory = AnalysisDirectory(".")
        else:
            if arguments.version:
                binary_version = get_binary_version_from_file(
                    arguments.local_configuration
                )
                log.stdout.write(
                    "binary version: {}\nclient version: {}".format(
                        binary_version, __version__
                    )
                )
                return ExitCode.SUCCESS

            configuration = Configuration(
                local_configuration=arguments.local_configuration,
                search_path=arguments.search_path,
                binary=arguments.binary,
                typeshed=arguments.typeshed,
                preserve_pythonpath=arguments.preserve_pythonpath,
                excludes=arguments.exclude,
            )
            if configuration.disabled:
                LOG.log(
                    log.SUCCESS, "Pyre will not run due to being explicitly disabled"
                )
                return ExitCode.SUCCESS

            if arguments.command in [commands.Kill]:
                analysis_directory = AnalysisDirectory(".")
            else:
                isolate = (
                    arguments.command in [commands.Analyze, commands.Check]
                    and not arguments.use_global_shared_analysis_directory
                )
                analysis_directory = resolve_analysis_directory(
                    arguments, commands, configuration, isolate=isolate
                )

        command = arguments.command
        exit_code = (
            # pyre-fixme: need to disentangle the initialize command.
            command(arguments, configuration, analysis_directory)
            .run()
            .exit_code()
        )
    except buck.BuckException as error:
        LOG.error(str(error))
        if arguments.command == commands.Persistent:
            commands.Persistent.run_null_server(timeout=3600 * 12)
        exit_code = ExitCode.BUCK_ERROR
    except EnvironmentException as error:
        LOG.error(str(error))
        if arguments.command == commands.Persistent:
            commands.Persistent.run_null_server(timeout=3600 * 12)
        exit_code = ExitCode.FAILURE
    except commands.ClientException as error:
        LOG.error(str(error))
        exit_code = ExitCode.FAILURE
    except Exception as error:
        LOG.error(str(error))
        LOG.info(traceback.format_exc())
        exit_code = ExitCode.FAILURE
    except KeyboardInterrupt:
        LOG.warning("Interrupted by user")
        LOG.debug(traceback.format_exc())
        exit_code = ExitCode.SUCCESS
    finally:
        log.cleanup(arguments)
        if analysis_directory:
            analysis_directory.cleanup()
        if configuration and configuration.logger:
            log_statistics(
                "perfpipe_pyre_usage",
                arguments=arguments,
                configuration=configuration,
                integers={
                    "exit_code": exit_code,
                    "runtime": int((time.time() - start) * 1000),
                },
                normals={"cwd": os.getcwd()},
            )

    return exit_code


if __name__ == "__main__":
    try:
        os.getcwd()
    except FileNotFoundError:
        print(
            "Pyre could not determine the current working directory. "
            "Has it been removed?\nExiting."
        )
        sys.exit(ExitCode.FAILURE)
    sys.exit(main())



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe


from .command import Command
from .incremental import Incremental
from .start import Start
from .stop import Stop


class Restart(Command):
    NAME = "restart"

    def __init__(self, arguments, configuration, analysis_directory) -> None:
        super(Restart, self).__init__(arguments, configuration, analysis_directory)

    def _run(self) -> None:
        Stop(self._arguments, self._configuration, self._analysis_directory).run()
        # Force the incremental run to be blocking.
        # pyre-fixme[16]: `Namespace` has no attribute `nonblocking`.
        self._arguments.nonblocking = False
        Incremental(
            self._arguments, self._configuration, self._analysis_directory
        ).run()



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import argparse
import enum
import logging
import os
import re
import resource
import signal
import subprocess
import threading
from abc import abstractmethod
from typing import Iterable, List, Optional, Set  # noqa

from .. import log
from ..configuration import Configuration
from ..exceptions import EnvironmentException
from ..filesystem import AnalysisDirectory, make_pyre_directory, remove_if_exists


LOG = logging.getLogger(__name__)  # type: logging.Logger


class ClientException(Exception):
    pass


class State(enum.IntEnum):
    DEAD = 0
    RUNNING = 1


class ExitCode(enum.IntEnum):
    SUCCESS = 0
    FOUND_ERRORS = 1
    FAILURE = 2
    BUCK_ERROR = 3
    # If the process exited due to a signal, this will be the negative signal number.
    SIGSEGV = -signal.SIGSEGV


class IncrementalStyle(enum.Enum):
    SHALLOW = "shallow"
    TRANSITIVE = "transitive"
    FINE_GRAINED = "fine_grained"

    def __str__(self) -> str:
        return self.value


class Result:
    def __init__(self, code: int, output: str) -> None:
        self.code = code
        self.output = output  # type: str

    def check(self) -> None:
        if self.code != ExitCode.SUCCESS:
            description = ":\n{}".format(self.output) if self.output else ""
            if self.code == ExitCode.SIGSEGV:
                description += (
                    "\nThis is a Pyre bug. Please re-run Pyre with --debug "
                    "and provide the output to the developers."
                )
            raise ClientException(
                "Client exited with error code {}{}".format(self.code, description)
            )


def typeshed_search_path(typeshed_root: str) -> List[str]:
    search_path = []
    typeshed_subdirectories = ["stdlib", "third_party"]
    for typeshed_subdirectory_name in typeshed_subdirectories:
        typeshed_subdirectory = os.path.join(typeshed_root, typeshed_subdirectory_name)
        if (
            not os.path.isdir(typeshed_subdirectory)
            or typeshed_subdirectory_name == "tests"
            or typeshed_subdirectory_name[0] == "."
        ):
            continue

        # Always prefer newer version over older version
        version_names = reversed(sorted(os.listdir(typeshed_subdirectory)))
        for version_name in version_names:
            # Anything under 2/ or 2.x is unusable for Pyre
            if version_name.startswith("2") and version_name != "2and3":
                continue
            search_path.append(os.path.join(typeshed_subdirectory, version_name))
    return search_path


# pyre-fixme[44]: `Command` non-abstract class with abstract methods
class Command:
    _buffer = []  # type: List[str]
    _call_client_terminated = False  # type: bool

    _exit_code = ExitCode.SUCCESS  # type: ExitCode

    _local_root = ""  # type: str

    def __init__(
        self,
        arguments: argparse.Namespace,
        configuration: Configuration,
        analysis_directory: AnalysisDirectory,
    ) -> None:
        self._arguments = arguments
        self._configuration = configuration

        self._analysis_directory = analysis_directory
        self._debug = arguments.debug  # type: bool
        self._enable_profiling = arguments.enable_profiling  # type: bool
        self._sequential = arguments.sequential  # type: bool
        self._strict = arguments.strict or (
            configuration and configuration.strict
        )  # type: bool
        self._additional_checks = arguments.additional_check  # type: List[str]
        self._show_error_traces = arguments.show_error_traces  # type: bool
        self._verbose = arguments.verbose  # type: bool
        self._hide_parse_errors = arguments.hide_parse_errors  # type: bool
        self._logging_sections = arguments.logging_sections  # type: str
        self._capable_terminal = arguments.capable_terminal  # type: bool
        self._log_identifier = arguments.log_identifier  # type: str
        self._logger = arguments.logger or (
            configuration and configuration.logger
        )  # type: str

        self._original_directory = arguments.original_directory  # type: str
        self._current_directory = arguments.current_directory  # type: str
        if arguments.local_configuration:
            self._local_root = (
                arguments.local_configuration
                if os.path.isdir(arguments.local_configuration)
                else os.path.dirname(arguments.local_configuration)
            )
        else:
            self._local_root = arguments.original_directory

    def run(self) -> "Command":
        self._run()
        return self

    def exit_code(self) -> int:
        return self._exit_code

    @abstractmethod
    def _run(self) -> None:
        """ Abstract method expected to be overridden by subclasses. """
        pass

    def _flags(self) -> List[str]:
        flags = []
        if self._debug:
            flags.extend(["-debug"])
        if self._sequential:
            flags.extend(["-sequential"])
        if self._strict:
            flags.extend(["-strict"])
        if self._additional_checks:
            flags.append("-additional-checks")
            flags.append(",".join(self._additional_checks))
        if self._show_error_traces:
            flags.append("-show-error-traces")
        if self._verbose:
            flags.append("-verbose")
        if not self._hide_parse_errors:
            if self._logging_sections:
                self._logging_sections = self._logging_sections + ",parser"
            else:
                self._logging_sections = "parser"
        if not self._capable_terminal:
            # Disable progress reporting for non-capable terminals.
            # This helps in reducing clutter.
            if self._logging_sections:
                self._logging_sections = self._logging_sections + ",-progress"
            else:
                self._logging_sections = "-progress"
        if self._logging_sections:
            flags.extend(["-logging-sections", self._logging_sections])
        if self._enable_profiling:
            pyre_directory = make_pyre_directory()
            profiling_output = os.path.join(pyre_directory, "profiling.log")
            # Clear the profiling log first since in pyre binary it's append-only
            remove_if_exists(profiling_output)
            flags.extend(["-profiling-output", profiling_output])
        if self._current_directory:
            flags.extend(["-project-root", self._current_directory])
        if self._log_identifier:
            flags.extend(["-log-identifier", self._log_identifier])
        if self._logger:
            flags.extend(["-logger", self._logger])
        return flags

    def _read_stdout(self, stdout: Iterable[bytes]) -> None:
        self._buffer = []
        for line in stdout:
            self._buffer.append(line.decode())

    def _read_stderr(self, stream: Iterable[bytes]) -> None:
        buffer = None
        log_pattern = re.compile(r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2} (\w+) (.*)")
        try:
            for line in stream:
                if self._call_client_terminated:
                    return
                line = line.decode().rstrip()
                match = log_pattern.match(line)
                if match:
                    if buffer:
                        buffer.flush()
                    buffer = log.Buffer(
                        section=match.groups()[0], data=[match.groups()[1]]
                    )
                elif buffer:
                    buffer.append(line)
            if buffer:
                buffer.flush()
        except Exception:
            pass

    def _call_client(self, command: str, capture_output: bool = True) -> Result:
        if not os.path.isdir(self._analysis_directory.get_root()):
            raise EnvironmentException(
                "`{}` is not a link tree.".format(self._analysis_directory.get_root())
            )

        client_command = [self._configuration.binary, command]
        client_command.extend(self._flags())
        client_command.append(self._analysis_directory.get_root())

        def limit_memory_usage() -> None:
            try:
                limit = 20 * 1024 * 1024 * 1024  # 20 GB
                resource.setrlimit(resource.RLIMIT_DATA, (limit, limit))
            except OSError:
                # Run the process with unlimited memory if the underlying syscall fails.
                pass

        LOG.debug("Running `%s`", " ".join(client_command))
        with subprocess.Popen(
            client_command,
            stdout=subprocess.PIPE if capture_output else None,
            stderr=subprocess.PIPE,
            preexec_fn=limit_memory_usage,
        ) as process:

            # Read stdout output
            if capture_output:
                stdout_reader = threading.Thread(
                    target=self._read_stdout, args=(process.stdout,)
                )
                stdout_reader.daemon = True
                stdout_reader.start()

            # Read the error output and print it.
            self._call_client_terminated = False
            stderr_reader = threading.Thread(
                target=self._read_stderr, args=(process.stderr,)
            )
            stderr_reader.daemon = True
            stderr_reader.start()

            # Wait for the process to finish and clean up.
            process.wait()
            self._call_client_terminated = True
            if capture_output:
                # pyre-fixme: stdout_reader is not always declared!
                stdout_reader.join()

            output = ""
            if capture_output:
                output = "".join(self._buffer)
            if process.returncode != 0 and capture_output:
                output = "".join(self._buffer)

            return Result(code=process.returncode, output=output)

    def _relative_path(self, path: str) -> str:
        return os.path.relpath(path, self._original_directory)

    def _state(self) -> State:
        pid_path = os.path.join(
            self._analysis_directory.get_root(), ".pyre/server/server.pid"
        )
        try:
            with open(pid_path) as file:
                pid = int(file.read())
                os.kill(pid, 0)  # throws if process is not running
            return State.RUNNING
        except Exception:
            return State.DEAD

    def _analysis_directory_string(self) -> str:
        return "`{}`".format(self._analysis_directory.get_root())



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

import errno
import fcntl
import functools
import logging
import os
import shutil
import subprocess
from contextlib import contextmanager
from time import time
from typing import Dict, Generator, Iterable, List, Optional, Set

from . import buck, log
from .exceptions import EnvironmentException


LOG = logging.getLogger(__name__)


class BuckBuilder:
    def build(self, targets: Iterable[str]) -> Iterable[str]:
        """
            Build the given targets, and return a list of output directories
            containing the target output.
        """
        raise NotImplementedError


def translate_path(root: str, path: str) -> str:
    if os.path.isabs(path):
        return path

    translated = os.path.join(root, path)
    if os.path.exists(translated):
        return os.path.realpath(translated)

    return path


def translate_paths(paths: Set[str], original_directory: str) -> Set[str]:
    current_directory = os.getcwd()
    if not original_directory.startswith(current_directory):
        return paths
    translation = os.path.relpath(original_directory, current_directory)
    if not translation:
        return paths
    return {translate_path(translation, path) for path in paths}


def find_root(original_directory: str, target_file: str) -> Optional[str]:
    current_directory = os.path.abspath(original_directory)
    while current_directory != "/":
        absolute = os.path.join(current_directory, target_file)
        if os.path.isfile(absolute):
            return current_directory
        current_directory = os.path.dirname(current_directory)
    return None


def exists(path: str) -> str:
    if not os.path.isfile(path):
        raise ValueError("%s is not a valid file" % path)
    return path


def is_parent(parent: str, child: str) -> bool:
    return child.startswith(parent.rstrip(os.sep) + os.sep)


def make_pyre_directory() -> str:
    pyre_directory = ".pyre"
    try:
        os.mkdir(pyre_directory)
    except FileExistsError:
        pass
    return pyre_directory


class UpdatedPaths:
    def __init__(self, updated: List[str]) -> None:
        self.updated = updated

    def is_empty(self) -> bool:
        return not self.updated


class AnalysisDirectory:
    def __init__(
        self,
        path: str,
        filter_paths: Optional[List[str]] = None,
        search_path: Optional[List[str]] = None,
    ) -> None:
        self._path = path
        self._filter_paths = filter_paths or []
        self._search_path = search_path or []

    def get_root(self) -> str:
        return self._path

    def get_pyre_server_directory(self) -> str:
        return os.path.join(self.get_root(), ".pyre/server")

    def get_filter_root(self) -> List[str]:
        return self._filter_paths or [self.get_root()]

    def prepare(self) -> None:
        pass

    def compute_symbolic_links(self) -> Dict[str, str]:
        return {}

    def process_updated_files(self, paths: List[str]) -> UpdatedPaths:
        """
            Process a list of paths which were added/removed/updated, making any
            necessary changes to the directory:
                - For an AnalysisDirectory, nothing needs to be changed, since
                  the mapping from source file to analysis file is 1:1.
                - For a SharedAnalysisDirectory, the symbolic links (as well as
                  the reverse-mapping we track) need to be updated to account for
                  new and deleted files.

            Return a list of files (corresponding to the given paths) that Pyre
            should be tracking.
        """
        tracked_paths = [path for path in paths if self._is_tracked(path)]
        return UpdatedPaths(updated=tracked_paths)

    def cleanup(self) -> None:
        pass

    @property
    @functools.lru_cache(1)
    def _tracked_directories(self) -> List[str]:
        tracked_directories = [
            self.get_root(),
            *[os.path.join(*path.split("$")) for path in self._search_path],
        ]
        return [os.path.abspath(path) for path in tracked_directories]

    def _is_tracked(self, path: str) -> bool:
        return any(
            is_parent(directory, path) for directory in self._tracked_directories
        )


class SharedAnalysisDirectory(AnalysisDirectory):
    def __init__(
        self,
        source_directories: List[str],
        targets: List[str],
        original_directory: Optional[str] = None,
        filter_paths: Optional[List[str]] = None,
        local_configuration_root: Optional[str] = None,
        extensions: Optional[List[str]] = None,
        search_path: Optional[List[str]] = None,
        isolate: bool = False,
        buck_builder: Optional[BuckBuilder] = None,
    ):
        self._source_directories = set(source_directories)
        self._targets = set(targets)
        self._original_directory = original_directory
        self._filter_paths = filter_paths or []
        self._local_configuration_root = local_configuration_root
        self._extensions = set(extensions or []) | {"py", "pyi"}
        self._search_path = search_path or []
        self._isolate = isolate
        self._buck_builder = buck_builder or buck.SimpleBuckBuilder()

        # Mapping from source files in the project root to symbolic links in the
        # analysis directory.
        self._symbolic_links = {}  # type: Dict[str, str]

    def get_scratch_directory(self) -> str:
        try:
            return (
                subprocess.check_output(["scratch", "path", "--subdir", "pyre"])
                .decode("utf-8")
                .strip()
            )
        except Exception:
            return os.path.join(os.getcwd(), ".pyre")

    @functools.lru_cache(1)
    def get_root(self) -> str:
        path_to_root = self._local_configuration_root or "shared_analysis_directory"
        suffix = "_{}".format(str(os.getpid())) if self._isolate else ""
        return os.path.join(
            self.get_scratch_directory(), "{}{}".format(path_to_root, suffix)
        )

    def get_filter_root(self) -> List[str]:
        return self._filter_paths or [os.getcwd()]

    # Exposed for testing.
    def _resolve_source_directories(self):
        if self._targets:
            new_source_directories = self._buck_builder.build(self._targets)
            original_directory = self._original_directory
            if original_directory is not None:
                new_source_directories = translate_paths(
                    new_source_directories, original_directory
                )
            self._source_directories.update(new_source_directories)

        if len(self._source_directories) == 0:
            raise EnvironmentException("No targets or source directories to analyze.")

    def prepare(self) -> None:
        start = time()
        root = self.get_root()
        LOG.info("Constructing shared directory `%s`", root)

        self._resolve_source_directories()

        try:
            os.makedirs(root)
        except OSError:
            pass  # Swallow.

        lock = os.path.join(root, ".pyre.lock")
        with acquire_lock(lock, blocking=True):
            self._clear()
            self._merge()
            LOG.log(
                log.PERFORMANCE, "Merged analysis directories in %fs", time() - start
            )
        self._symbolic_links.update(self.compute_symbolic_links())

    def compute_symbolic_links(self) -> Dict[str, str]:
        return _compute_symbolic_link_mapping(self.get_root(), self._extensions)

    def process_updated_files(self, paths: List[str]) -> UpdatedPaths:
        """
            Return the paths in the analysis directory (symbolic links)
            corresponding to the given paths.
            Result also includes any files which are within a tracked directory.

            This method will remove/add symbolic links for deleted/new files.
        """
        tracked_files = []
        deleted_paths = [path for path in paths if not os.path.isfile(path)]
        new_paths = [
            path
            for path in paths
            if path not in self._symbolic_links
            and os.path.isfile(path)
            and is_parent(os.getcwd(), path)
        ]
        updated_paths = [
            path
            for path in paths
            if path not in deleted_paths and path not in new_paths
        ]

        for path in updated_paths:
            if path in self._symbolic_links:
                tracked_files.append(self._symbolic_links[path])
            elif self._is_tracked(path):
                tracked_files.append(path)

        return UpdatedPaths(updated=tracked_files)

    def cleanup(self) -> None:
        try:
            if self._isolate:
                shutil.rmtree(self.get_root())
        except Exception:
            pass

    def _clear(self):
        root = self.get_root()
        for path in os.listdir(root):
            if path.startswith(".pyre"):
                continue

            path = os.path.join(root, path)
            remove_if_exists(path)

    def _merge(self) -> None:
        root = self.get_root()

        all_paths = {}
        for source_directory in self._source_directories:
            self._merge_into_paths(source_directory, all_paths)
        for relative, original in all_paths.items():
            merged = os.path.join(root, relative)
            add_symbolic_link(merged, original)

    # Exposed for testing.
    def _merge_into_paths(
        self, source_directory: str, all_paths: Dict[str, str]
    ) -> None:
        paths = find_python_paths(root=source_directory)
        for path in paths:
            relative = os.path.relpath(path, source_directory)
            if not path:
                continue
            # don't bother stat'ing paths that are already in the analysis directory.
            if relative in all_paths:
                continue
            try:
                absolute = os.path.realpath(path)
                # Don't merge symlinked directories.
                if not os.path.isfile(absolute):
                    continue
                if relative.endswith("__init__.py") and is_empty(absolute):
                    # Don't let empty __init__.py files override legitimate files.
                    continue
                all_paths[relative] = absolute
            except FileNotFoundError:
                continue


def find_paths_with_extensions(root: str, extensions: Iterable[str]) -> List[str]:
    root = os.path.abspath(root)  # Return absolute paths.
    extension_filter = []
    for extension in extensions:
        if len(extension_filter) > 0:
            extension_filter.append("-or")
        extension_filter.extend(["-name", "*.{}".format(extension)])

    output = (
        subprocess.check_output(
            [
                "find",
                root,
                # All files ending with the given extensions ...
                "(",
                *extension_filter,
                ")",
                # ... and that are either regular files ...
                "(",
                "-type",
                "f",
                "-or",
                # ... or symlinks.
                "-type",
                "l",
                ")",
                # Print all such files.
                "-print",
            ],
            stderr=subprocess.DEVNULL,
        )
        .decode("utf-8")
        .strip()
    )
    return output.split("\n") if output else []


def find_python_paths(root: str) -> List[str]:
    try:
        return find_paths_with_extensions(root, ["py", "pyi"])
    except subprocess.CalledProcessError:
        raise EnvironmentException(
            "Pyre was unable to locate an analysis directory. "
            "Ensure that your project is built and re-run pyre."
        )


def is_empty(path: str) -> bool:
    try:
        return os.stat(path).st_size == 0
    except FileNotFoundError:
        return False


def remove_if_exists(path: str) -> None:
    try:
        os.remove(path)
    except OSError:
        pass  # Not a file.
    try:
        shutil.rmtree(path)
    except OSError:
        pass  # Not a directory.


def _compute_symbolic_link_mapping(
    directory: str, extensions: Iterable[str]
) -> Dict[str, str]:
    """
        Given a shared analysis directory, produce a mapping from actual source files
        to files contained within this directory. Only includes files which have
        one of the provided extensions.

        Watchman watches actual source files, so when a change is detected to a
        file, this mapping can be used to identify what file changed from Pyre's
        perspective.
    """
    symbolic_links = {}
    try:
        for symbolic_link in find_paths_with_extensions(directory, extensions):
            symbolic_links[os.path.realpath(symbolic_link)] = symbolic_link
    except subprocess.CalledProcessError as error:
        LOG.warning(
            "Exception encountered trying to find source files "
            "in the analysis directory: `%s`",
            error,
        )
        LOG.warning("Starting with an empty set of tracked files.")
    return symbolic_links


def _delete_symbolic_link(link_path: str) -> None:
    os.unlink(link_path)


def add_symbolic_link(link_path: str, actual_path: str) -> None:
    directory = os.path.dirname(link_path)
    try:
        os.makedirs(directory)
    except OSError:
        pass
    try:
        os.symlink(actual_path, link_path)
    except OSError as error:
        if error.errno == errno.EEXIST:
            os.unlink(link_path)
            os.symlink(actual_path, link_path)
        else:
            LOG.error(str(error))


@contextmanager
def acquire_lock(path: str, blocking: bool) -> Generator[Optional[int], None, None]:
    """Raises an OSError if the lock can't be acquired"""
    try:
        with open(path, "w+") as lockfile:
            if not blocking:
                lock_command = fcntl.LOCK_EX | fcntl.LOCK_NB
            else:
                lock_command = fcntl.LOCK_EX

            fcntl.lockf(lockfile.fileno(), lock_command)
            yield lockfile.fileno()
            fcntl.lockf(lockfile.fileno(), fcntl.LOCK_UN)

    except FileNotFoundError:
        yield


class Filesystem:
    def list(
        self, root: str, patterns: List[str], exclude: Optional[List[str]] = None
    ) -> List[str]:
        """
            Return the list of files that match any of the patterns within root.
            If exclude is provided, files that match an exclude pattern are omitted.

            Note: The `find` command does not understand globs properly.
                e.g. 'a/*.py' will match 'a/b/c.py'
            For this reason, avoid calling this method with glob patterns.
        """

        command = ["find", "."]
        command += self._match_any(patterns)
        if exclude:
            command += ["-and", "!"]
            command += self._match_any(exclude)
        return (
            subprocess.run(command, stdout=subprocess.PIPE, cwd=root)
            .stdout.decode("utf-8")
            .split()
        )

    def _match_any(self, patterns: List[str]) -> List[str]:
        expression = []
        for pattern in patterns:
            if expression:
                expression.append("-or")
            expression.extend(["-path", "./{}".format(pattern)])
        return ["(", *expression, ")"]


class MercurialBackedFilesystem(Filesystem):
    def list(
        self, root: str, patterns: List[str], exclude: Optional[List[str]] = None
    ) -> List[str]:
        try:
            command = ["hg", "files"]
            for pattern in patterns:
                command += ["--include", pattern]
            if exclude:
                for pattern in exclude:
                    command += ["--exclude", pattern]
            return (
                subprocess.run(
                    command, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, cwd=root
                )
                .stdout.decode("utf-8")
                .split()
            )
        except FileNotFoundError:
            raise EnvironmentException("hg executable not found.")


@functools.lru_cache(1)
def get_filesystem() -> Filesystem:
    try:
        subprocess.check_output(["hg", "status"], stderr=subprocess.DEVNULL)
        return MercurialBackedFilesystem()
    except (subprocess.CalledProcessError, FileNotFoundError):
        return Filesystem()



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe
from typing import List

from .. import get_binary_version
from ..version import __version__
from .command import Command


class Rage(Command):
    NAME = "rage"

    def __init__(self, arguments, configuration, analysis_directory) -> None:
        super(Rage, self).__init__(arguments, configuration, analysis_directory)
        # pyre-fixme[16]: `Namespace` has no attribute `command`.
        self._arguments.command = self.NAME
        self._configuration = configuration

    def _flags(self) -> List[str]:
        return []

    def _run(self) -> None:
        # Do not use logging. Logging goes to stderr.
        print("Client version:", __version__, flush=True)
        print("Binary path:", self._configuration.binary, flush=True)
        print(
            "Configured binary version:",
            get_binary_version(self._configuration),
            flush=True,
        )
        self._call_client(command=self.NAME, capture_output=False).check()



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# pyre-strict

import argparse
import logging
import os
import shutil
import signal
import subprocess
from typing import Iterable  # noqa

import psutil

from .. import BINARY_NAME, CLIENT_NAME
from ..configuration import Configuration
from ..filesystem import AnalysisDirectory
from .command import Command


LOG = logging.getLogger(__name__)  # type: logging.Logger


class Kill(Command):
    NAME = "kill"

    def __init__(
        self,
        arguments: argparse.Namespace,
        configuration: Configuration,
        analysis_directory: AnalysisDirectory,
    ) -> None:
        super(Kill, self).__init__(arguments, configuration, analysis_directory)

    def _run(self) -> None:
        # Kills all processes that have the same binary as the one specified
        # in the configuration.
        binary_name = _get_process_name("PYRE_BINARY", BINARY_NAME)
        subprocess.run(["pkill", binary_name])
        server_root = os.path.join(
            os.path.realpath(self._analysis_directory.get_root()), ".pyre", "server"
        )

        def _delete_linked_path(link_path: str) -> None:
            try:
                actual_path = os.readlink(link_path)
                os.remove(actual_path)
            except OSError:
                pass
            try:
                os.unlink(link_path)
            except OSError:
                pass

        _delete_linked_path(os.path.join(server_root, "server.sock"))
        _delete_linked_path(os.path.join(server_root, "json_server.sock"))

        if self._arguments.with_fire is True:
            # If a resource cache exists, delete it to remove corrupted artifacts.
            try:
                shutil.rmtree(os.path.join(os.getcwd(), ".pyre/resource_cache"))
            except OSError:
                pass
            # If a buck builder cache exists, also remove it.
            try:
                shutil.rmtree("/tmp/pyre/buck_builder_cache")
            except OSError:
                pass

        for process in psutil.process_iter(attrs=["name"]):
            if process.info["name"] != CLIENT_NAME:
                continue
            # We need to be careful about how we kill the client here, as otherwise we
            # might cause a race where we attempt to kill the `pyre kill` command.
            pid_to_kill = process.pid
            if pid_to_kill == os.getpgid(os.getpid()):
                continue
            try:
                os.kill(pid_to_kill, signal.SIGKILL)
            except ProcessLookupError:
                continue


def _get_process_name(environment_variable_name: str, default: str) -> str:
    overridden = os.getenv(environment_variable_name)
    if overridden is not None:
        return os.path.basename(overridden)
    else:
        return default



# Copyright (c) 2019-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import argparse
import functools
import logging
import os
import socket
import subprocess
import sys
from typing import Any, BinaryIO, Dict, Iterable, List, Optional, Set  # noqa

from . import language_server_protocol
from .configuration import Configuration
from .filesystem import AnalysisDirectory, find_root
from .watchman_subscriber import Subscription, WatchmanSubscriber


LOG = logging.getLogger(__name__)  # type: logging.Logger


class MonitorException(Exception):
    pass


class SocketConnection(object):
    def __init__(self, socket_path: str) -> None:
        self.socket = socket.socket(
            socket.AF_UNIX, socket.SOCK_STREAM
        )  # type: socket.socket
        self.socket.connect(socket_path)
        self.input = self.socket.makefile(mode="rb")  # type: BinaryIO
        self.output = self.socket.makefile(mode="wb")  # type: BinaryIO

    def close(self) -> None:
        try:
            self.socket.close()
        except OSError:
            pass

    def __del__(self) -> None:
        """
            Monitor is created and then runs in a forked process.
            As a result, in the original process, this object gets garbage
            collected without the socket being closed.
            For this reason, we explicitly close the socket on destruction.
        """
        self.close()


class ProjectFilesMonitor(WatchmanSubscriber):
    """
        Logs from this monitor are found in .pyre/file_monitor/file_monitor.log
        One file monitor is spawned per pyre server. When a server is stopped,
        the process of pyre file monitor associated to it is killed.
    """

    NAME = "file_monitor"

    def __init__(
        self,
        arguments: argparse.Namespace,
        configuration: Configuration,
        analysis_directory: AnalysisDirectory,
    ) -> None:
        super(ProjectFilesMonitor, self).__init__(analysis_directory)
        self._arguments = arguments
        self._configuration = configuration
        self._analysis_directory = analysis_directory

        self._extensions = set(
            ["py", "pyi"] + configuration.extensions
        )  # type: Set[str]

        self._watchman_path = self._find_watchman_path(
            arguments.current_directory
        )  # type: str

        socket_path = os.path.join(
            self._analysis_directory.get_root(), ".pyre", "server", "json_server.sock"
        )
        self._socket_connection = self._connect_to_socket(
            socket_path
        )  # type: SocketConnection

        try:
            language_server_protocol.perform_handshake(
                self._socket_connection.input,
                self._socket_connection.output,
                self._configuration.version_hash,
            )
        except (OSError, ValueError) as error:
            raise MonitorException(
                "Exception encountered during handshake: `{}`".format(error)
            )

    @property
    def _name(self) -> str:
        return self.NAME

    @property
    @functools.lru_cache(1)
    def _subscriptions(self) -> List[Subscription]:
        subscription = {
            "expression": [
                "allof",
                ["type", "f"],
                ["not", "empty"],
                ["anyof", *[["suffix", extension] for extension in self._extensions]],
            ],
            "fields": ["name"],
        }
        return [
            Subscription(
                self._watchman_path, "pyre_file_change_subscription", subscription
            )
        ]

    @staticmethod
    def pid_path(analysis_directory_root: str) -> str:
        return os.path.join(
            analysis_directory_root,
            ".pyre",
            ProjectFilesMonitor.NAME,
            "{}.pid".format(ProjectFilesMonitor.NAME),
        )

    @staticmethod
    def is_alive(analysis_directory_root: str) -> bool:
        pid_path = ProjectFilesMonitor.pid_path(analysis_directory_root)
        try:
            with open(pid_path) as file:
                pid = int(file.read())
                os.kill(pid, 0)  # throws if process is not running
            return True
        except Exception:
            return False

    def _handle_response(self, response: Dict[str, Any]) -> None:
        try:
            absolute_paths = [
                os.path.join(response["root"], path) for path in response["files"]
            ]
            LOG.info("Received Watchman update for files %s.", absolute_paths)

            updated_paths = self._analysis_directory.process_updated_files(
                absolute_paths
            )

            if updated_paths.is_empty():
                LOG.info("Skipping update: Pyre doesn't track any of these files.")
                return

            LOG.info("Notifying server of update to files %s.", updated_paths.updated)
            message = language_server_protocol.LanguageServerProtocolMessage(
                method="updateFiles",
                parameters={"files": updated_paths.updated, "invalidated": []},
            )
            if not language_server_protocol.write_message(
                self._socket_connection.output, message
            ):
                LOG.info("Failed to communicate with server. Shutting down.")
                self._alive = False  # terminate daemon
                self._socket_connection.close()
        except KeyError:
            pass

    @staticmethod
    def _find_watchman_path(directory: str) -> str:
        watchman_path = find_root(directory, ".watchmanconfig")
        if not watchman_path:
            raise MonitorException(
                "Could not find a watchman directory from "
                "the current directory `{}`".format(directory)
            )
        return watchman_path

    @staticmethod
    def _connect_to_socket(socket_path: str) -> SocketConnection:
        try:
            return SocketConnection(os.path.realpath(socket_path))
        except (ConnectionRefusedError, FileNotFoundError, OSError) as error:
            raise MonitorException(
                "Failed to connect to server at `{}`. Reason: `{}`".format(
                    socket_path, error
                )
            )



# Copyright (c) 2016-present, Facebook, Inc.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import argparse
from typing import List, Optional  # noqa

from .. import log
from ..configuration import Configuration
from ..filesystem import AnalysisDirectory
from .check import Check


class Analyze(Check):
    NAME = "analyze"  # type: str

    def __init__(
        self,
        arguments: argparse.Namespace,
        configuration: Configuration,
        analysis_directory: AnalysisDirectory,
    ) -> None:
        super(Analyze, self).__init__(arguments, configuration, analysis_directory)
        self._analysis = arguments.analysis  # type: str
        self._no_verify = arguments.no_verify  # type: bool
        self._taint_models_path = (
            arguments.taint_models_path or configuration.taint_models_path
        )  # type: List[str]
        self._save_results_to = arguments.save_results_to  # type: Optional[str]
        self._dump_call_graph = arguments.dump_call_graph  # type: bool

    def _flags(self) -> List[str]:
        flags = super()._flags()
        flags.extend(["-analysis", self._analysis])
        if self._taint_models_path:
            for path in self._taint_models_path:
                flags.extend(["-taint-models", path])
        save_results_to = self._save_results_to
        if save_results_to:
            flags.extend(["-save-results-to", save_results_to])
        if self._dump_call_graph:
            flags.append("-dump-call-graph")
        if self._no_verify:
            flags.append("-no-verify")
        return flags

    def _run(self, retries: int = 1) -> None:
        self._analysis_directory.prepare()
        result = self._call_client(command=self.NAME)
        result.check()
        log.stdout.write(result.output)
